/*
   CScanner - A Chinese Lexical Scanner
   Copyright (C) 2000-2001 Chih-Hao Tsai

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
 */

/*
   Module: cscanner.c
   Version: 0.26
   Date: April 7, 2001
   Author: Chih-Hao Tsai
   Email: hao520@yahoo.com
   CScanner website: http://technology.chtsai.org/cscanner/
   Author's homepage: http://chtsai.org/
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "cscanner.h"

#define CHINESE_WORD_LENGTH_MAX 8
#define DEFAULT_LEXICON "tsi.src"	/* Any word list with the word as
					   the first token of each line will do */


/* Character sets (used internally by c_scanner_get_token) */
typedef enum
{
  UNKNOWN, ASCII, BIG5
}
Characterset;

/* BIG5 handling functions */
unsigned int big5_code_to_sernum (unsigned char, unsigned char);
int big5_char_test (unsigned int);	/* returns 1 if normal characters, 0 if symbols or undefined */

/* Character frequency table (defined in charfreq.c) */
extern unsigned char big5_character_freq[13060];

/* Character frequency lookup helper */
static int lookup_big5_charfreq (unsigned char, unsigned char,
				 const Lexicon *);

/* Implementation of simple forward matching and complex maximum
   matching algorithm described in "MMSEG: A Word Identification
   System for Mandarin Chinese Text Based on Two Variants of the
   Maximum Matching Algorithm"
   http://www.geocities.com/hao510/mmseg/
 */

typedef struct weight Weight;
struct weight
{
  int n_word[2];
  int word[2][3];
  int sum_len[2];
  float mean_len[2];
  float ss[2];
  float variance[2];
  int sum_charfreq[2];
};

static void *skiplists_duplicates_helper (int *, int *);
static int fmm_get_first_word (char *, const Lexicon *);
static int mmseg_get_first_word (char *, const Lexicon *);
static void mmseg_helper (char *, const Lexicon *, Weight *, int *, int, int);
static void mmseg_weight_helper (char *, const Lexicon *, Weight *,
				 int *, int);
static void mmseg_promote_helper (Weight *);


Lexicon *
c_scanner_load_lexicon (const char *filename)
{
  Lexicon *L;
  FILE *WORDLIST;
  char *tok, *word;
  int freq, *freq_ptr;
  char string[256];		/* Maximum line size of the input file */

  L = (Lexicon *) malloc (sizeof (Lexicon));
  if (!L)
    return NULL;

  L->size = 0;
  L->charfreq = (unsigned char *) big5_character_freq;
  L->Word = skiplists_new ((SkiplistsCompareFunc) strcmp);
  skiplists_disallow_duplicates (L->Word,
				 (SkiplistsDuplicateFunc)
				 skiplists_duplicates_helper);

  if (!filename || filename[0] == '\0')
    filename = DEFAULT_LEXICON;

  if ((WORDLIST = fopen (filename, "rt")) == NULL)
    {
      skiplists_destroy (L->Word);
      free (L);
      return NULL;
    }

  while (fgets (string, 255, WORDLIST))
    if ((word = strtok (string, " \n\t")))
      if (strlen (word) >= 2 && strlen (word) <= CHINESE_WORD_LENGTH_MAX * 2)
	if (is_big5_chinese_char (*(word), *(word + 1)))
	  {
	    tok = strtok (0, " \n\t");
	    if (tok)
	      freq = atoi (tok);
	    else
	      freq = 0;
	    freq_ptr = (int *) malloc (sizeof (int));
	    *freq_ptr = freq;
	    if (skiplists_insert (L->Word, strdup (word), freq_ptr))
	      L->size++;
	  }

  fclose (WORDLIST);

  return L;
}


static void *
skiplists_duplicates_helper (int *old_value, int *new_value)
{
  (*old_value) += *new_value;
  return (old_value);
}



void
c_scanner_free_lexicon (Lexicon * L)
{
  skiplists_destroy (L->Word);
  free (L);
}


void *
c_scanner_search_lexicon (char *key, const Lexicon * L)
{
  return skiplists_lookup_with_finger (L->Word, key);
}


CScanner *
c_scanner_new (Lexicon * L)
{
  CScanner *scanner;
  if (!L)
    return NULL;

  scanner = (CScanner *) malloc (sizeof (CScanner));
  if (!scanner)
    return NULL;

  memset (scanner, 0, sizeof (CScanner));
  scanner->token_length_max = TOKEN_LENGTH_MAX;
  scanner->chinese_word_length_max = CHINESE_WORD_LENGTH_MAX;
  scanner->lexicon = L;
  scanner->type = UNRECOGNIZED;
  return scanner;
}


int
c_scanner_input_text (CScanner * scanner, const char *text, int length)
{
  if (!scanner || !text || text[0] == '\0')
    return 0;

  scanner->text = text;
  scanner->length = (length > 0
		     && length < strlen (text)) ? length : strlen (text);
  scanner->token[0] = '\0';
  scanner->type = UNRECOGNIZED;
  scanner->token_position = 0;
  scanner->position = 0;
  return 1;
}


void
c_scanner_free (CScanner * scanner)
{
  free (scanner);
}

char *
c_scanner_get_token (CScanner * scanner, int complexity)
{
  int got_a_token = 0, tokenlen = 0;
  int chinese_word_scansize;
  unsigned int p = scanner->position, cursor = 0;
  unsigned char *buffer;
  Characterset mode = UNKNOWN;
  unsigned char c, c1;

  chinese_word_scansize = complexity == 0 ?
    (scanner->chinese_word_length_max - 1) * 2 :
    (scanner->chinese_word_length_max * 3 - 1) * 2;

  buffer =
    (unsigned char *) malloc ((scanner->token_length_max + 1) *
			      sizeof (unsigned char));
  *(buffer) = 0;

  while (!got_a_token && ((p + cursor) < scanner->length)
	 && cursor < scanner->token_length_max)
    {
      c = *(scanner->text + p + cursor);
      if (iscntrl (c) || isspace (c))	/* skipping white spaces */
	{
	  cursor++;
	  if (strlen (buffer) > 0)
	    got_a_token = 1;
	  else
	    {
	      p += cursor;
	      cursor = 0;
	    }
	}
      else
	{
	  if (mode == UNKNOWN)	/* first contact of a token */
	    {
	      *(buffer + cursor++) = c;
	      *(buffer + cursor) = 0;
	      if (isgraph (c))
		{
		  mode = ASCII;
		  if (ispunct (c))
		    {
		      scanner->type = ASCII_SYMBOL;
		      got_a_token = 1;
		    }
		  else
		    scanner->type = ASCII_WORD;
		}
	      else
		{
		  if (is_big5_leading_char (c))
		    {
		      if ((p + cursor) < scanner->length)
			{
			  c1 = *(scanner->text + p + cursor);
			  if (is_big5_trailing_char (c1))
			    {
			      mode = BIG5;
			      *(buffer + cursor++) = c1;
			      *(buffer + cursor) = 0;
			      if (is_big5_symbol (c, c1))
				{
				  scanner->type = BIG5_SYMBOL;
				  got_a_token = 1;
				}
			      else if (is_big5_undefined_char (c, c1))
				{
				  scanner->type = BIG5_UNDEFINED;
				  got_a_token = 1;
				}
			      else if (is_big5_chinese_char (c, c1))
				{
				  scanner->type = BIG5_WORD;
				}
			    }
			  else
			    {
			      mode = ASCII;
			      scanner->type = ASCII_8BIT;
			      got_a_token = 1;
			    }
			}
		      else
			{
			  mode = ASCII;
			  scanner->type = ASCII_8BIT;
			  got_a_token = 1;
			}
		    }
		  else
		    {
		      mode = ASCII;
		      scanner->type = ASCII_8BIT;
		      got_a_token = 1;
		    }
		}
	    }

	  else if (mode == ASCII)	/* nth contact of an ASCII token */
	    {
	      if (isalnum (c))
		{
		  *(buffer + cursor++) = c;
		  *(buffer + cursor) = 0;
		}
	      else
		{
		  cursor--;
		  got_a_token = 1;
		}
	    }
	  else if (mode == BIG5)	/* nth contact of a BIG5 token */
	    {
	      if (is_big5_leading_char (c))
		{
		  if (((p + cursor + 1) < scanner->length)
		      && (cursor <= chinese_word_scansize))
		    {
		      c1 = *(scanner->text + p + cursor + 1);
		      if (is_big5_trailing_char (c1))
			{
			  if (is_big5_chinese_char (c, c1))
			    {
			      *(buffer + cursor++) = c;
			      *(buffer + cursor++) = c1;
			      *(buffer + cursor) = 0;
			    }
			  else
			    got_a_token = 1;
			}
		      else
			got_a_token = 1;
		    }
		  else
		    got_a_token = 1;
		}
	      else
		got_a_token = 1;
	    }
	  else
	    got_a_token = 1;
	}
    }
  if (got_a_token || strlen (buffer) >= 1)	/* Got a token or end of line */
    {
      if (mode == ASCII)
	{
	  if (scanner->type == ASCII_SYMBOL || scanner->type == ASCII_8BIT)
	    {
	      strcpy (scanner->token, buffer);
	      scanner->token_position = p;
	      scanner->position = p + cursor;
	      free (buffer);
	      return scanner->token;
	    }
	  else
	    {
	      tokenlen = strlen (buffer);
	      while (ispunct (*(buffer + tokenlen - 1)) && tokenlen > 0)
		tokenlen--;	/* remove trailing punctuation marks */
	      *(buffer + tokenlen) = 0;
	      strcpy (scanner->token, buffer);
	      scanner->token_position = p;
	      scanner->position = p + tokenlen;
	      free (buffer);
	      return scanner->token;
	    }
	}
      else if (mode == BIG5)
	{
	  if (scanner->type == BIG5_SYMBOL || scanner->type == BIG5_UNDEFINED)
	    {
	      strcpy (scanner->token, buffer);
	      scanner->token_position = p;
	      scanner->position = p + cursor;
	      free (buffer);
	      return scanner->token;
	    }
	  else
	    {

	      tokenlen = strlen (buffer);
	      if (tokenlen == 2)	/* one character */
		{
		  strcpy (scanner->token, buffer);
		  scanner->token_position = p;
		  scanner->position = p + cursor;
		  free (buffer);
		  return scanner->token;
		}
	      else
		{
		  if (complexity == 0)
		    /* forward maximum matching */
		    tokenlen = fmm_get_first_word (buffer, scanner->lexicon);
		  else
		    /* complex maximum matching */
		    tokenlen =
		      mmseg_get_first_word (buffer, scanner->lexicon);
		  strncpy (scanner->token, buffer, tokenlen);
		  *(scanner->token + tokenlen) = 0;
		  scanner->token_position = p;
		  scanner->position = p + tokenlen;
		  free (buffer);
		  return scanner->token;
		}
	    }
	}
      else
	{
	  free (buffer);
	  return NULL;
	}
    }
  free (buffer);
  return NULL;
}


static int
fmm_get_first_word (char *buffer, const Lexicon * L)
{
  int tokenlen;

  tokenlen = strlen (buffer);

  if (tokenlen <= 0)
    return 0;
  else if (tokenlen <= 2)
    return tokenlen;
  while (tokenlen > 2)
    {
      if (c_scanner_search_lexicon (buffer, L))
	break;
      else
	{
	  tokenlen -= 2;
	  *(buffer + tokenlen) = 0;
	}
    }
  return tokenlen;
}

static int
mmseg_get_first_word (char *buffer, const Lexicon * L)
{
  int tokenlen;
  int w[3] = { 0, 0, 0 };
  Weight *W;

  tokenlen = strlen (buffer);

  if (tokenlen <= 0)
    return 0;
  else if (tokenlen <= 2)
    return tokenlen;

  W = (Weight *) malloc (sizeof (Weight));
  memset (W, 0, sizeof (Weight));
  W->variance[0] = 99;

  mmseg_helper (buffer, L, W, w, 0, 0);

  tokenlen = W->word[0][0] * 2;
  free (W);
  return tokenlen;
}

static void
mmseg_helper (char *buffer, const Lexicon * L, Weight * W, int *w,
	      int position, int count)
{
  int i, blen, range;
  unsigned char temp[CHINESE_WORD_LENGTH_MAX * 2 + 1];

  blen = strlen (buffer + position) / 2;
  range = blen < CHINESE_WORD_LENGTH_MAX ? blen : CHINESE_WORD_LENGTH_MAX;

  if (count == 2 || range <= 0)
    mmseg_weight_helper (buffer, L, W, w, count);
  else
    for (i = 1; i <= range; i++)
      {
	strncpy (temp, buffer + position, i * 2);
	*(temp + i * 2) = 0;
	if (i == 1 || c_scanner_search_lexicon (temp, L))
	  {
	    w[count] = i;
	    mmseg_helper (buffer, L, W, w, position + i * 2, count + 1);
	  }
      }
}

static void
mmseg_weight_helper (char *buffer, const Lexicon * L, Weight * W,
		     int *w, int count)
{
  int i, j = 0;

  if (W && L && buffer && w && count > 0)
    {
      W->n_word[1] = count;
      W->sum_len[1] = 0;
      W->mean_len[1] = 0;
      W->ss[1] = 0;
      W->variance[1] = 0;
      W->sum_charfreq[1] = 0;

      for (i = 0; i < W->n_word[1]; i++)
	{
	  j += w[i];
	  W->word[1][i] = w[i];
	  W->sum_len[1] += W->word[1][i];
	  if (W->word[1][i] == 1)
	    W->sum_charfreq[1] +=
	      lookup_big5_charfreq (*(buffer + (j - 1) * 2),
				    *(buffer + (j - 1) * 2 + 1), L);
	}

      W->mean_len[1] = (float) W->sum_len[1] / (float) W->n_word[1];

      for (i = 0; i < count; i++)
	W->ss[1] +=
	  ((float) W->word[1][i] - W->mean_len[1]) * ((float) W->word[1][i] -
						      W->mean_len[1]);

      W->variance[1] = W->ss[1] / W->n_word[1];

      /*
         find maximal weight
       */
      if (W->sum_len[1] > W->sum_len[0])	/* Chen & Liu (1992) */
	mmseg_promote_helper (W);
      else if (W->sum_len[1] == W->sum_len[0])
	{
	  if (W->mean_len[1] > W->mean_len[0])	/* Tsai (1996) */
	    mmseg_promote_helper (W);
	  else if (W->mean_len[1] == W->mean_len[0])
	    {
	      if (W->variance[1] < W->variance[0])	/* Chen & Liu (1992) */
		mmseg_promote_helper (W);
	      else if (W->variance[1] == W->variance[0])
		{
		  if (W->sum_charfreq[1] > W->sum_charfreq[0])	/* Chen & Liu (1992) */
		    mmseg_promote_helper (W);
		}
	    }
	}
    }
}

static void
mmseg_promote_helper (Weight * W)
{
  if (W)
    {
      W->n_word[0] = W->n_word[1];
      W->word[0][0] = W->word[1][0];
      W->word[0][1] = W->word[1][1];
      W->word[0][2] = W->word[1][2];
      W->sum_len[0] = W->sum_len[1];
      W->mean_len[0] = W->mean_len[1];
      W->ss[0] = W->ss[1];
      W->variance[0] = W->variance[1];
      W->sum_charfreq[0] = W->sum_charfreq[1];

      W->n_word[1] = 0;
      W->word[1][0] = 0;
      W->word[1][1] = 0;
      W->word[1][2] = 0;
      W->sum_len[1] = 0;
      W->mean_len[1] = 0;
      W->ss[1] = 0;
      W->variance[1] = 0;
      W->sum_charfreq[1] = 0;
    }
}

int
lookup_big5_charfreq (unsigned char leading, unsigned char trailing,
		      const Lexicon * L)
{
  unsigned int p;

  if (leading && trailing && L)
    if (is_big5_chinese_char (leading, trailing))
      {
	p =
	  (unsigned int) (big5_code_to_sernum (leading, trailing) -
			  (unsigned int) 0x8800);
	return (int) *(L->charfreq + p);
      }
  return 0;
}

unsigned int
big5_code_to_sernum (unsigned char leading, unsigned char trailing)
{
  unsigned int sernum, serbase;

  if (leading == 0xc6 && trailing >= 0xa1)
    {
      serbase = 0x8001;
      leading -= 0xc6;
      trailing = trailing - 0xa1;
      sernum = leading * 0x9d + trailing + serbase;
      return sernum;
    }
  if (leading >= 0xa1 && leading <= 0xa3)
    {
      serbase = 0x8400;
      leading -= 0xa1;
    }
  else if (leading >= 0xa4 && leading <= 0xc6)
    {
      serbase = 0x8800;
      leading -= 0xa4;
    }
  else if (leading >= 0xc7 && leading <= 0xc8)
    {
      serbase = 0x805f;
      leading -= 0xc7;
    }
  else if (leading >= 0xc9 && leading <= 0xf9)
    {
      serbase = 0x9d19;
      leading -= 0xc9;
    }
  else
    return 0;

  if (trailing >= 0x40 && trailing <= 0x7e)
    trailing -= 0x40;
  else if (trailing >= 0xa1 && trailing <= 0xfe)
    trailing = trailing - 0xa1 + 0x3f;
  else
    return 0;

  sernum = leading * 0x9d + trailing + serbase;
  return sernum;
}

int
is_big5_leading_char (unsigned char leading)
{
  if (!leading)
    return 0;
  else if (leading >= 0xa1 && leading <= 0xf9)
    return 1;
  else
    return 0;
}

int
is_big5_trailing_char (unsigned char trailing)
{
  if (!trailing)
    return 0;
  else if ((trailing >= 0x40 && trailing <= 0x7e)
	   || (trailing >= 0xa1 && trailing <= 0xfe))
    return 1;
  else
    return 0;
}

int
big5_char_test (unsigned int sernum)
{
  if (sernum >= 0x8001 && sernum <= 0x816d)
    return 0;
  else if (sernum >= 0x8400 && sernum <= 0x8597)
    return 0;
  else if (sernum >= 0x8800 && sernum <= 0xBB03)
    return 1;
  else if (sernum >= 0xBB04 && sernum <= 0xBB25)
    return 0;
  else
    return -1;
}

int
is_big5_chinese_char (unsigned char leading, unsigned char trailing)
{
  if (leading && trailing)
    return (big5_char_test (big5_code_to_sernum (leading, trailing)) > 0);
  else
    return 0;
}

int
is_big5_symbol (unsigned char leading, unsigned char trailing)
{
  if (leading && trailing)
    return (big5_char_test (big5_code_to_sernum (leading, trailing)) == 0);
  else
    return 0;
}

int
is_big5_undefined_char (unsigned char leading, unsigned char trailing)
{
  if (leading && trailing)
    return (big5_char_test (big5_code_to_sernum (leading, trailing)) < 0);
  else
    return 0;
} 
